@@ -4,9 +4,15 @@ SELECT
 FROM
     Author;
 
+--! authors_stream()#
+SELECT
+    *
+FROM
+    Author;
+
 --! books() {title}*
 SELECT
-    Title::asd
+    Title
 FROM
     Book;
 
